from vit_flax.layers import Transformer

from flax import nn
from jax import lax


class ViT(nn.Module):
  """Vision transformer

  :param x: Input tensor image
  :param patch_size: Patch dimension from image
  :param dim: Latent dim
  :param depth: Number of layers of Residual-normalized attention layers.
  :param num_heads: Number of attention heads
  :param dense_dims: Tuple(int, int) - (Transformer FC dim, Classifier FC dim) 
  :param img_size: Dimension of input image
  :param num_classes: Number of classification classes
  :param initializer: Flax initializer

  :return:Classification output
  """
  def apply(self, x,
            patch_size, dim, depth, num_heads, dense_dims,
            img_size, num_classes,
            initializer=nn.initializers.normal(stddev=1.0)):
    b, h, w, c = x.shape
    patch = x.reshape(b, c, h, w)
    num_patches = (img_size // patch_size) ** 2
    patch = patch.reshape(b, (h*w)//(patch_size*patch_size), c*patch_size**2)

    fc_embedding = nn.Dense(patch, dim)

    class_tokens = self.param('class_tokens', (b, 1, dim), initializer)
    pos_embedding = self.param(
      'pos_embedding', (1, num_patches+1, dim), initializer
    )
    x = lax.concatenate([class_tokens, fc_embedding], dimension=1)
    x += pos_embedding
    x = Transformer(x, depth, num_heads, dense_dims[0])

    x = x[:, 0]
    x = nn.Dense(x, dense_dims[1])
    x = nn.gelu(x)
    x = nn.Dense(x, num_classes)
    return x



from flax import nn


class FeedForward(nn.Module):
  """Simple FeedForward module

  x -> Dense(x, latent_dim) -> GeLU -> Dense(x, original_dim)
  """
  def apply(self, x, latent_dim):
    dim = x.shape[-1]
    x = nn.Dense(x, latent_dim)
    x = nn.gelu(x)
    x = nn.Dense(x, dim)
    return x


class Residual(nn.Module):
  """Simple residual function

  x -> fn(x) + x"""
  def apply(self, x, residual_fn):
    return residual_fn(x) + x


class PreNorm(nn.Module):
  """Apply function to a normalized modulue
  """
  def apply(self, x, norm, fn):
    return fn(norm(x))


class Transformer(nn.Module):
  """A simple implementation of a Transformer

  :param x: Input tensor.
  :param depth: Number of layers of Residual-normalized attention layers.
  :param num_heads: Number of attention heads
  :param feed_forward_dim: FC dimension

  :return: Transformer output embedding
  """
  def apply(self, x, depth, num_heads, feed_forward_dim_1):
    attention = nn.SelfAttention.partial(num_heads=num_heads)
    norm = nn.LayerNorm
    norm_attention = PreNorm.partial(norm=norm, fn=attention)
    residual_attention = Residual.partial(residual_fn=norm_attention)

    forward = FeedForward.partial(latent_dim=feed_forward_dim_1)
    norm_forward = PreNorm.partial(norm=norm, fn=forward)
    residual_forward = Residual.partial(residual_fn=norm_forward)

    for _ in range(depth):
      x = residual_forward(residual_attention(x))

    return x
